import collections
import configparser
import errno
import os
import re

import aiobtclientapi
from xdg.BaseDirectory import xdg_config_home

from . import __project_name__, _errors

from . import _debug  # noqa:F401 isort:skip


DEFAULT_CONFIG_FILEPATH = os.path.join(xdg_config_home, __project_name__, 'config.cfg')
DEFAULT_LOCATIONS_FILEPATH = os.path.join(xdg_config_home, __project_name__, 'locations')
DEFAULT_CLIENTS_FILEPATH = os.path.join(xdg_config_home, __project_name__, 'clients.ini')


def _as_dict(cfg):
    # Convert ConfigParser object to dict
    return {
        section: dict(cfg.items(section))
        for section in cfg.sections()
    }


def _convert_to_client(options, option):
    client_name = options[option]
    if client_name not in aiobtclientapi.client_names():
        raise ValueError(f'{client_name}: Unknown client')
    return client_name


def _convert_to_url(options, option):
    url = options[option]
    client_name = options['client']
    client_cls = getattr(aiobtclientapi, f'{client_name.capitalize()}API')
    try:
        return client_cls.URL(url)
    except ValueError as e:
        raise ValueError(f'{url}: {e}')


_bool_values = {
    '1': True, 'yes': True, 'true': True, 'on': True,
    '0': False, 'no': False, 'false': False, 'off': False,
}

def _convert_to_bool(options, option):
    value = str(options[option]).lower()
    try:
        return _bool_values[value]
    except KeyError:
        raise ValueError(f'{value}: Must be true/false, yes/no, on/off, 1/0')


def _convert_to_string(options, option):
    return str(options[option]).replace('\n', ' ')


def _convert_to_list(options, option):
    value = str(options[option]).strip()
    if value:
        return value.split('\n')
    else:
        return []


def _convert_to_umask(options, option):
    value = options[option]
    if value is None:
        return None
    elif not value:
        return 0
    elif re.search(r'^[0-7]{1,4}$', str(value)):
        return int(value, 8)
    else:
        raise ValueError(f'{value}: Invalid umask')


class Config(collections.abc.Mapping):
    """
    :class:`dict` subclass that reads simple `option = value` lines from file

    Values may contain line breaks if there is some whitespace after each line
    break. Use :func:`_convert_to_list` or :func:`_convert_to_string` to handle
    line breaks.

    :param filepath: File to read configuration from

    :raise ConfigError: if reading or parsing `filepath` fails
    """

    def __init__(self, filepath):
        self._filepath = filepath
        self._config = self._read(filepath)

    def _read(self, filepath):

        try:
            text = open(filepath, 'r').read()
        except OSError as e:
            # Ignore missing default config file path
            if e.errno != errno.ENOENT or filepath != DEFAULT_CONFIG_FILEPATH:
                msg = e.strerror if e.strerror else str(e)
                raise _errors.ConfigError(f'Failed to read: {msg}', filepath=filepath)
            else:
                text = ''

        # Validation
        dct = self._parse(text, filepath)
        self._ensure_only_valid_options(dct, filepath)
        self._fill_in_defaults(dct)
        self._convert_values(dct, filepath)
        return dct

    def _parse(self, text, filepath):
        dct = {}
        current_option = None

        for i, line in enumerate(re.split('[\n\r]', text), start=1):
            # comment or empty line
            if re.search(r'^\s*#', line) or not line.strip():
                continue

            # option = value
            match = re.search(r'^\s*(\w+)\s*=\s*(.*?)\s*$', line)
            if match:
                dct[match.group(1)] = match.group(2)
                current_option = match.group(1)
                continue

            # option = value1\n value2
            # or
            # option = \n value1\n value2
            match = re.search(r'^\s+(.*?)\s*$', line)
            if match and current_option is not None:
                if match.group(1).strip():
                    if dct[current_option]:
                        dct[current_option] += '\n' + match.group(1)
                    else:
                        dct[current_option] = match.group(1)
                continue

            raise _errors.ConfigError(f'{line}: Invalid syntax',
                                      filepath=filepath, line_number=i)

        return dct

    _valid_options = (
        'before_location_search_commands',
        'after_location_found_commands',
        'after_torrent_handled_commands',
        'umask',
    )

    def _ensure_only_valid_options(self, dct, filepath):
        for option in dct:
            if option not in self._valid_options:
                raise _errors.ConfigError(f'Unknown option: {option}', filepath=filepath)

    _defaults = {
        'before_location_search_commands': '',
        'after_location_found_commands': '',
        'after_torrent_handled_commands': '',
        'umask': None,
    }

    def _fill_in_defaults(self, dct):
        for option in self._defaults:
            if option not in dct:
                dct[option] = self._defaults[option]

    _convert_values_map = {
        'before_location_search_commands': _convert_to_list,
        'after_location_found_commands': _convert_to_list,
        'after_torrent_handled_commands': _convert_to_list,
        'umask': _convert_to_umask,
    }

    def _convert_values(self, dct, filepath):
        for option, converter in self._convert_values_map.items():
            try:
                dct[option] = converter(dct, option)
            except ValueError as e:
                raise _errors.ConfigError(f'{option}: {e}', filepath=filepath)

    @property
    def filepath(self):
        """Where the configuration was read from"""
        return self._filepath

    def __getitem__(self, key):
        return self._config[key]

    def __len__(self):
        return len(self._config)

    def __iter__(self):
        return iter(self._config)

    def __repr__(self):
        return f'<{type(self).__name__} {self._filepath!r} {self._config!r}>'


class Locations(collections.abc.MutableSequence):
    """
    :class:`list` subclass that reads directory paths from `filepath`

    :param locations: Directory paths
    :param filepath: File to read more `locations` from

    The format of `filepath` is very simple:

        * Directory paths are separated by newlines.

        * Lines that start with "#" and empty lines are ignored.

        * If a path ends with ``f"{os.sep}*"``, all subdirectories in that
          directory are added.

        * Basic environment variable expansion is supported, e.g. "$HOME/foo".

    :raise ConfigError: if reading `filepath` fails or if it contains a file
        path
    """

    def __init__(self, *locations, filepath):
        self._filepath = filepath
        self._list = []
        # self.extend() should normalize `locations`. _read() does that on its
        # own and provides the proper filepath and line_number to ConfigError.
        self.extend(locations)
        self._list.extend(self._read(filepath))

    @property
    def filepath(self):
        return self._filepath

    def __repr__(self):
        return f'<{type(self).__name__} {self._filepath!r} {self._list!r}>'

    def _read(self, filepath):
        locations = []

        try:
            with open(filepath, 'r') as f:
                for line_number, line in enumerate(f.readlines(), start=1):
                    line = line.strip()
                    if line and not line.startswith('#'):
                        for normalized_path in self._normalize(line, filepath, line_number):
                            if normalized_path not in locations:
                                locations.append(normalized_path)

        except OSError as e:
            # Ignore missing default config file path
            if e.errno == errno.ENOENT and filepath == DEFAULT_LOCATIONS_FILEPATH:
                pass
            else:
                msg = e.strerror if e.strerror else str(e)
                raise _errors.ConfigError(f'Failed to read: {msg}', filepath=filepath)

        return locations

    @classmethod
    def _normalize(cls, line, filepath, line_number):
        subdirs = []

        if line.endswith(f'{os.sep}*'):
            # Expand "*"
            parent_dir = line[:-2]
            try:
                subdir_names = os.listdir(parent_dir)
            except OSError as e:
                msg = e.strerror if e.strerror else str(e)
                raise _errors.ConfigError(f'Failed to read subdirectories from {parent_dir}: {msg}',
                                          filepath=filepath, line_number=line_number)
            else:
                for name in subdir_names:
                    subdir_path = os.path.join(parent_dir, name)
                    # Exclude non-directories (files and exotic stuff like
                    # sockets), but include nonexisting paths (subdir_path may
                    # contain unresolved environment variables and download
                    # locations may not exist anyway)
                    if os.path.isdir(subdir_path) or not os.path.exists(subdir_path):
                        subdirs.append(subdir_path)
        else:
            subdirs.append(line)

        # Resolve environment variables
        subdirs = [
            cls._resolve_env_vars(subdir, filepath, line_number)
            for subdir in subdirs
        ]

        # Complain if subdir exists but is not a directory
        for subdir in subdirs:
            if os.path.exists(subdir) and not os.path.isdir(subdir):
                raise _errors.ConfigError(f'Not a directory: {subdir}',
                                          filepath=filepath, line_number=line_number)

        return sorted(subdirs)

    @classmethod
    def _resolve_env_vars(cls, line, filepath, line_number):
        # Resolve "~/foo" and "~user/foo"
        path = os.path.expanduser(line)

        while True:
            # Find valid variable name
            # https://stackoverflow.com/a/2821201
            match = re.search(r'\$([a-zA-Z_]+[a-zA-Z0-9_]*)', path)
            if not match:
                break
            else:
                env_var_name = match.group(1)
                env_var_value = os.environ.get(env_var_name, None)
                if env_var_value is None:
                    raise _errors.ConfigError(f'Unset environment variable: ${env_var_name}',
                                              filepath=filepath, line_number=line_number)
                elif env_var_value == '':
                    raise _errors.ConfigError(f'Empty environment variable: ${env_var_name}',
                                              filepath=filepath, line_number=line_number)
                else:
                    path = path.replace(f'${env_var_name}', env_var_value)

        return path

    def __setitem__(self, index, value):
        # `index` can be int or slice and `value` can be one path or list of
        # paths.
        if not isinstance(value, str) and isinstance(value, collections.abc.Iterable):
            normalized_paths = []
            for item in value:
                for normalized_path in self._normalize(item, None, None):
                    if normalized_path not in self._list:
                        normalized_paths.append(normalized_path)
            self._list[index] = normalized_paths

        else:
            normalized_path = self._normalize(value, None, None)[0]
            if normalized_path not in self._list:
                self._list[index] = normalized_path

    def insert(self, index, value):
        normalized_paths = self._normalize(value, None, None)
        self._list.insert(index, normalized_paths[0])

    def __getitem__(self, index):
        return self._list[index]

    def __delitem__(self, index):
        del self._list[index]

    def __len__(self):
        return len(self._list)

    def __eq__(self, other):
        if isinstance(other, collections.abc.Sequence):
            return self._list == list(other)
        else:
            return NotImplemented


class Clients(collections.abc.Mapping):
    """
    :class:`dict` subclass that reads BitTorrent client configs from INI file

    :param filepath: File to read client configurations from

    :raise ConfigError: if reading or parsing `filepath` fails
    """

    def __init__(self, filepath):
        self._filepath = filepath
        self._config = self._read(filepath)

    def _read(self, filepath):
        cfg = configparser.ConfigParser(
            default_section=None,
            interpolation=None,
            allow_no_value=False,
            delimiters=('=',),
            comment_prefixes=('#',),
        )
        try:
            cfg.read_string(
                open(filepath, 'r').read(),
                source=filepath,
            )
        except OSError as e:
            # Ignore missing default config file path
            if e.errno != errno.ENOENT or filepath != DEFAULT_CLIENTS_FILEPATH:
                msg = e.strerror if e.strerror else str(e)
                raise _errors.ConfigError(f'Failed to read: {msg}', filepath=filepath)
        except configparser.MissingSectionHeaderError as e:
            raise _errors.ConfigError(f'Line {e.lineno}: {e.line.strip()}: Option outside of section', filepath=filepath)
        except configparser.ParsingError as e:
            lineno, msg = e.errors[0]
            raise _errors.ConfigError(f'Line {lineno}: {msg.strip()}: Invalid syntax', filepath=filepath)
        except configparser.DuplicateSectionError as e:
            raise _errors.ConfigError(f'Line {e.lineno}: {e.section}: Duplicate section', filepath=filepath)
        except configparser.DuplicateOptionError as e:
            raise _errors.ConfigError(f'Line {e.lineno}: {e.option}: Duplicate option', filepath=filepath)
        except configparser.Error as e:
            raise _errors.ConfigError(str(e), filepath=filepath)

        # Validation
        dct = _as_dict(cfg)
        self._ensure_only_valid_options(dct, filepath)
        self._ensure_mandatory_options(dct, filepath)
        self._fill_in_defaults(dct)
        self._convert_values(dct, filepath)
        return dct

    _valid_options = (
        'client',
        'url',
        'username',
        'password',
        'verify',
        'stopped',
    )

    def _ensure_only_valid_options(self, dct, filepath):
        for section in dct:
            for option in dct[section]:
                if option not in self._valid_options:
                    raise _errors.ConfigError(f'{section}: Unknown option: {option}', filepath=filepath)

    _mandatory_options = ('client',)

    def _ensure_mandatory_options(self, dct, filepath):
        for section in dct:
            for option in self._mandatory_options:
                if option not in dct[section]:
                    raise _errors.ConfigError(f'{section}: Missing option: {option}', filepath=filepath)

    _defaults = {
        # Empty URL means "default URL" when passed to one of the
        # aiobtclientapi.*API.URL classes (see _convert_to_url())
        'url': '',
        'username': '',
        'password': '',
        'verify': False,
        'stopped': False,
    }

    # Client-specific defaults
    _client_defaults = collections.defaultdict(
        lambda: collections.defaultdict(lambda: None),
        {
            'transmission': collections.defaultdict(
                lambda: None,
                {
                    # Transmission can't add torrents without verifying
                    'verify': True,
                },
            ),
        },
    )

    def _fill_in_defaults(self, dct):
        for section in dct:
            for option in self._defaults:
                if option not in dct[section]:
                    client_name = dct[section]['client']
                    client_default_value = self._client_defaults[client_name][option]
                    if client_default_value is not None:
                        dct[section][option] = client_default_value
                    else:
                        dct[section][option] = self._defaults[option]

    def _convert_values(self, dct, filepath):
        convert_map = {
            'client': _convert_to_client,
            'url': _convert_to_url,
            'username': _convert_to_string,
            'password': _convert_to_string,
            'verify': _convert_to_bool,
            'stopped': _convert_to_bool,
        }
        for section, options in dct.items():
            for option, converter in convert_map.items():
                try:
                    options[option] = converter(options, option)
                except ValueError as e:
                    raise _errors.ConfigError(f'{section}: {option}: {e}', filepath=filepath)

    @property
    def filepath(self):
        """Where the configuration was read from"""
        return self._filepath

    @property
    def default(self):
        """First client config in the file"""
        section = next(iter(self._config))
        return self._config[section]

    def __getitem__(self, key):
        return self._config[key]

    def __len__(self):
        return len(self._config)

    def __iter__(self):
        return iter(self._config)

    def __repr__(self):
        return f'<{type(self).__name__} {self._filepath!r} {self._config!r}>'



from functools import cached_property

import aiobtclientapi


class Client:
    def __init__(self, config):
        self._config = config.copy()

    @property
    def config(self):
        return self._config.copy()

    @cached_property
    def api(self):
        """:class:`aiobtclientapi.APIBase` subclass instance"""
        return aiobtclientapi.api(
            name=self._config['client'],
            url=self._config['url'],
            username=self._config['username'],
            password=self._config['password'],
        )

    async def add_torrent(self, torrent, location):
        """Add `torrent` at download directory `location`"""
        async with self.api:
            return await self.api.add(
                torrent,
                location=location,
                verify=self._config['verify'],
                stopped=self._config['stopped'],
            )



import collections


class Combinator(collections.abc.Iterable):
    """
    Take a dictionary that maps keys to lists and turn it into a list of all
    possible ``(key, list item)`` combinations

    Example:

    >>> things = {
    >>>     'a': ['a1', 'a2'],
    >>>     'b': ['b1'],
    >>>     'c': ['c1', 'c2', 'c3', 'c4'],
    >>> }
    >>> list(_Combinator(things))
    >>> [
    >>>     [('a', 'a1'), ('b', 'b1'), ('c', 'c1')],
    >>>     [('a', 'a1'), ('b', 'b1'), ('c', 'c2')],
    >>>     [('a', 'a1'), ('b', 'b1'), ('c', 'c3')],
    >>>     [('a', 'a1'), ('b', 'b1'), ('c', 'c4')],
    >>>     [('a', 'a2'), ('b', 'b1'), ('c', 'c1')],
    >>>     [('a', 'a2'), ('b', 'b1'), ('c', 'c2')],
    >>>     [('a', 'a2'), ('b', 'b1'), ('c', 'c3')],
    >>>     [('a', 'a2'), ('b', 'b1'), ('c', 'c4')],
    >>> ]
    """

    def __init__(self, things):
        self._things = {
            key: tuple(items)
            for key, items in things.items()
        }
        self._keys = tuple(self._things)
        self._indexes = {
            key: 0 if len(self._things[key]) > 0 else -1
            for key in self._keys
        }
        self._locked_keys = set()

    @property
    def _current_pairs(self):
        # List of (key, list item) tuples for each key using current list
        # indexes.
        pairs = []
        for key in self._keys:
            item = self._things[key][self._indexes[key]]
            pairs.append((key, item))
        return pairs

    def __iter__(self):
        # Take the first step so the while loop below doesn't immediately end
        # because all indexes are zero.
        # Do not take the first step if any list is empty.
        if (
            all(index >= 0 for key, index in self._indexes.items())
            and any(key not in self._locked_keys for key in self._keys)
        ):
            yield self._current_pairs
            self._advance()

        # Yield over pairs until all keys are either at their start position 0
        # again or locked to a specific index. Don't loop at all if any list is
        # empty (indicated by index == -1, see __init__).
        while any(index > 0 and key not in self._locked_keys
                  for key, index in self._indexes.items()):
            yield self._current_pairs
            self._advance()

    def _advance(self):
        # Increase the rightmost index or reset it to zero and increase the next
        # index on the left.
        for key in reversed(self._keys):
            if key not in self._locked_keys:
                if self._indexes[key] < len(self._things[key]) - 1:
                    self._indexes[key] += 1
                    break
                else:
                    self._indexes[key] = 0
            else:
                continue

    def lock(self, *keys):
        """
        Do not change the current combination for every `key` in `keys`

        You should call this if you are happy with the current `key` ->
        `list_item` combination to prevent iterating over any other combinations
        for `key`. This can safe lots of iterations.

        :param keys`: Any key from the `things` mapping from initialization
        """
        for key in keys:
            if key not in self._keys:
                raise RuntimeError(f'Cannot exclude unknown key: {key!r}')
        self._locked_keys.update(keys)



__project_name__ = 'tofipa'
__description__ = 'Get download directory from torrent file'
__homepage__ = 'https://codeberg.org/plotski/tofipa'
__version__ = '0.3.0'
__author__ = 'plotski'
__author_email__ = 'plotski@example.org'

import logging  # isort:skip
from . import __project_name__  # isort:skip
_debug = logging.getLogger(__project_name__).debug

from ._cli import cli
from ._errors import FindError
from ._location import FindDownloadLocation



import argparse
import asyncio
import os
import subprocess
import sys

import torf

from . import (__description__, __project_name__, __version__, _btclient,
               _config, _errors, _location)


def _parse_args(args):
    argparser = argparse.ArgumentParser(
        prog=__project_name__,
        description=__description__,
    )
    argparser.add_argument(
        'TORRENT',
        nargs='+',
        help='Path to torrent file',
    )
    argparser.add_argument(
        '--config-file',
        default=_config.DEFAULT_CONFIG_FILEPATH,
        help='File containing configuration options',
    )
    argparser.add_argument(
        '--location', '-l',
        # TODO: Always use "extend" when Python 3.7 is no longer supported.
        #       https://docs.python.org/3/library/argparse.html#action
        action='extend' if sys.version_info >= (3, 8, 0) else 'append',
        nargs='+' if sys.version_info >= (3, 8, 0) else None,
        default=[],
        help=(
            'Potential download location of existing files in TORRENT '
            '(may be given multiple times)'
        ),
    )
    argparser.add_argument(
        '--locations-file',
        default=_config.DEFAULT_LOCATIONS_FILEPATH,
        help='File containing newline-separated list of download locations',
    )
    argparser.add_argument(
        '--default', '-d',
        help='Default location if no existing files are found',
    )
    argparser.add_argument(
        '--clients-file',
        default=_config.DEFAULT_CLIENTS_FILEPATH,
        help='File containing BitTorrent client connections',
    )
    argparser.add_argument(
        '--client', '-c',
        default=None,
        help='Add TORRENT to CLIENT (CLIENT is a section name in CLIENTS_FILE)',
    )
    argparser.add_argument(
        '--noclient', '-C',
        action='store_true',
        default=None,
        help='Do not add TORRENT to any client, print download location instead',
    )
    argparser.add_argument(
        '--version',
        action='version',
        version=f'{__project_name__} {__version__}',
    )
    argparser.add_argument(
        '--debug',
        nargs='?',
        metavar='FILE',
        help='Write debugging messages to FILE or STDERR if FILE is "-"',
    )
    return argparser.parse_args(args)


def _fatal_error(msg):
    sys.stderr.write(f'{msg}\n')
    sys.exit(1)


def _setup_debugging(args):
    # Debugging
    if args.debug == '-':
        import logging
        logging.basicConfig(stream=sys.stderr, level=logging.DEBUG)
    elif args.debug:
        import logging
        logging.basicConfig(filename=args.debug, level=logging.DEBUG)


def _get_config(args):
    # Read main configuration from file
    try:
        return _config.Config(filepath=args.config_file)
    except _errors.ConfigError as e:
        _fatal_error(str(e))


def _get_locations(args):
    # Read locations file
    try:
        locations = _config.Locations(filepath=args.locations_file)
    except _errors.ConfigError as e:
        _fatal_error(str(e))
    else:
        # Prepend arguments from --location to locations from config file
        for location in args.location:
            if location in locations:
                locations.remove(location)
        locations[0:0] = args.location
        if not locations:
            _fatal_error(f'No locations specified. See: {__project_name__} --help')
        return locations


def _get_client_config(args):
    # Read client configurations from file
    try:
        configs = _config.Clients(filepath=args.clients_file)
    except _errors.ConfigError as e:
        _fatal_error(str(e))
    else:
        if args.client is not None:
            try:
                return configs[args.client]
            except KeyError:
                _fatal_error(f'Unknown client: {args.client}')
        elif configs:
            return configs.default


def _get_client(args):
    if not args.noclient:
        config = _get_client_config(args)
        if config:
            return _btclient.Client(config)


def _find_location(torrent, args, config):
    location_finder = _location.FindDownloadLocation(
        torrent=torrent,
        locations=_get_locations(args),
        default=args.default,
        umask=config['umask'],
    )
    try:
        return location_finder.find()
    except _errors.FindError as e:
        _fatal_error(str(e))


async def _handle_location(location, torrent, client):
    if client:
        response = await client.add_torrent(torrent, location)
        for warning in response.warnings:
            sys.stderr.write(f'{torrent}: WARNING: {warning}\n')
        if response.errors:
            for error in response.errors:
                sys.stderr.write(f'{torrent}: {error}\n')
            return False  # Failure
        return True  # Success
    else:
        sys.stdout.write(f'{location}\n')
        return True  # Success


def _run_shell_commands(cmds, env):
    for cmd in cmds:
        try:
            subprocess.run(
                cmd,
                shell=True,
                check=True,
                env=env,
            )
        except subprocess.CalledProcessError as e:
            _fatal_error(f'Command terminated with exit code {e.returncode}: {cmd}')


def _get_torrent(torrent_path):
    try:
        return torf.Torrent.read(torrent_path)
    except torf.TorfError as e:
        _fatal_error(str(e))


async def _cli(args):
    args = _parse_args(args)
    _setup_debugging(args)
    config = _get_config(args)
    client = _get_client(args)

    exit_code = 0

    for torrent_path in args.TORRENT:
        torrent = _get_torrent(torrent_path)

        _run_shell_commands(config['before_location_search_commands'], env={
            'TOFIPA_TORRENT_FILE': str(torrent_path),
            'TOFIPA_TORRENT_NAME': str(torrent.name),
            'TOFIPA_TORRENT_LOCATION': '',
            'TOFIPA_TORRENT_PATH': '',
        })

        location = _find_location(torrent_path, args, config)

        _run_shell_commands(config['after_location_found_commands'], env={
            'TOFIPA_TORRENT_FILE': str(torrent_path),
            'TOFIPA_TORRENT_NAME': str(torrent.name),
            'TOFIPA_TORRENT_LOCATION': str(location),
            'TOFIPA_TORRENT_PATH': str(os.path.join(location, torrent.name)),
        })

        if not await _handle_location(location, torrent_path, client):
            exit_code = 1
        else:
            _run_shell_commands(config['after_torrent_handled_commands'], env={
                'TOFIPA_TORRENT_FILE': str(torrent_path),
                'TOFIPA_TORRENT_NAME': str(torrent.name),
                'TOFIPA_TORRENT_LOCATION': str(location),
                'TOFIPA_TORRENT_PATH': str(os.path.join(location, torrent.name)),
            })

    return exit_code


def cli():
    try:
        sys.exit(asyncio.run(_cli(sys.argv[1:])))
    except KeyboardInterrupt:
        _fatal_error('Cancelled')



import collections
import difflib
import errno
import functools
import os
import string
import tempfile

import torf

from . import __project_name__, _debug
from ._combinator import Combinator
from ._errors import FindError


class FindDownloadLocation:
    """
    Find download path for a torrent, linking existing files

    :param torrent: Path to torrent file
    :param locations: Sequence of directory paths that are searched for files in
        `torrent`
    :param default: Default download directory if none of the files in `torrent`
        are found
    :param umask: File mode creation mask to use when creating directories
        (e.g. `0o002`)

        https://en.wikipedia.org/wiki/umask

    :raise: :class:`~.FindDownloadLocation` if anything goes wrong
    """

    def __init__(self, *, torrent, locations, default=None, umask=None):
        self._torrent_filepath = str(torrent)
        self._torrent = None
        # Ugly hack to deduplicate locations while preserving order
        self._locations = tuple(
            dict.fromkeys(str(loc) for loc in locations)
        )
        if len(self._locations) < 1:
            raise RuntimeError('You must provide at least one potential download location')
        self._default_location = str(default) if default else None
        self._umask = umask
        self._found_files = set()

    def find(self):
        try:
            self._torrent = torf.Torrent.read(self._torrent_filepath)
        except torf.TorfError as e:
            raise FindError(e)
        else:
            if self._umask is not None:
                _debug('Using special umask: %r', self._umask)
                orig_umask = os.umask(self._umask)
            else:
                orig_umask = None

            try:
                download_location = self._get_download_location()
                if download_location is not None:
                    _debug('Found location: %r', download_location)
                    return download_location
                elif self._default_location:
                    _debug('Default location: %r', self._default_location)
                    return self._default_location
                else:
                    _debug('First location: %r', self._locations[0])
                    return self._locations[0]
            finally:
                if orig_umask is not None:
                    _debug('Restoring old umask: %r', orig_umask)
                    os.umask(orig_umask)

    def _get_download_location(self):
        candidates = self._get_size_matching_candidates()
        links_to_create = {}
        download_location = None

        # `paths` maps relative file paths expected by torrent to a candidate
        # dictionary from `candidates`.
        for paths in self._each_set_of_linked_candidates(candidates):
            _debug('%d / %d files', len(links_to_create), len(self._torrent.files))
            for file, candidate in paths.items():
                if file not in links_to_create:
                    if self._verify_file(file, location=candidate['temporary_location']):
                        # Use download location of first matching file.
                        if download_location is None:
                            _debug('Setting download location: %r', candidate['location'])
                            download_location = candidate['location']

                        _debug('  %s: Using %r', file, candidate['filepath'])
                        links_to_create[file] = (candidate['filepath'], os.path.join(download_location, file))

                        # Maintain a list of files we already found.
                        # _each_set_of_linked_candidates() uses that list to
                        # skip new combinations of files that are now
                        # irrelevant.
                        _debug('  Found files: %r', self._found_files)
                        self._found_files.add(file)

                    else:
                        _debug('  %s: Not using %r', file, candidate['filepath'])
                else:
                    _debug('  %s: Already found %r', file, links_to_create[file][0])

            if all(file in links_to_create for file in self._torrent.files):
                _debug('All files found: %r', tuple(links_to_create))
                break

        # Create final links to best matching files.
        for source, target in links_to_create.values():
            self._create_hardlink(source, target)

        return download_location

    def _each_set_of_linked_candidates(self, candidates):
        _debug('Iterating over combinations of temporary symlinks:')
        for file, cands in sorted(candidates.items()):
            _debug('  * %s -> %s', [c['filepath'] for c in cands], file)

        combinator = Combinator(candidates)
        for pairs in combinator:
            with self._temporary_directory as location:
                # Create temporary links as they are expected by the torrent.
                for file, candidate in pairs:
                    candidate['temporary_location'] = location
                    source = os.path.abspath(candidate['filepath'])
                    target = os.path.join(location, file)
                    self._create_symlink(source, target)

                # Map torrent file path to other relevant information that is
                # needed for matching.
                yield {file: candidate for file, candidate in pairs}

                # Do not iterate over new combinations of files we already
                # found.
                combinator.lock(*sorted(self._found_files))

            # Ensure temporary links are removed.
            assert not os.path.exists(location), location

    def _verify_file(self, file, location):
        _debug('  Verifying %s at %s', file, location)
        content_path = os.path.join(location, self._torrent.name)
        with torf.TorrentFileStream(self._torrent, content_path=content_path) as tfs:
            # Don't check the first and the last piece of a file as they likely
            # overlap with another file that might be either invalid or missing.
            file_piece_indexes = tfs.get_absolute_piece_indexes(file, (1, -2))
            _debug('    Verifying pieces: %r', file_piece_indexes)
            for piece_index in file_piece_indexes:
                piece_ok = tfs.verify_piece(piece_index)
                if piece_ok is True:
                    _debug('    Piece %d is valid', piece_index)
                elif piece_ok is False:
                    _debug('    Piece %d is invalid', piece_index)
                    return False
                elif piece_ok is None:
                    _debug('    Piece %d is unverifiable; probably non-existing file', piece_index)
                    return None
        return True

    def _get_size_matching_candidates(self):
        # Map each relative file path from the torrent to a list of candidates.
        # A candidate is a dictionary that stores information about a file from
        # the file system that has the expected size.
        candidates = collections.defaultdict(lambda: [])
        torrent_files = tuple(self._torrent.files)
        for filepath, location in self._each_file(*self._locations):
            for file in torrent_files:
                if self._is_size_match(file, filepath):
                    filepath_rel = filepath[len(location):].lstrip(os.sep)

                    def get_similarity(filepath_fs, filepath_torrent=str(file)):
                        return self._get_path_similarity(filepath_fs, filepath_torrent)

                    # Candidates are dictionaries:
                    #         file: torf.File object (relative file path in torrent)
                    #     location: Download path to pass to the BitTorrent
                    #               client along with the torrent file
                    #     filepath: Existing file path with same size as `file`
                    # filepath_rel: `filepath` without `location`
                    #               (`location` / `filepath` is the same as `filepath`)
                    #   similarity: How close `file` is to `filepath_rel` as
                    #               float from 0.0 to 1.0
                    candidates[file].append({
                        'location': location,
                        'filepath': filepath,
                        'filepath_rel': filepath_rel,
                        'similarity': get_similarity(filepath_rel),
                    })

        # Sort size-matching files by file path similarity.
        for file in candidates:
            candidates[file].sort(key=lambda c: c['similarity'], reverse=True)

            _debug('Size matches for %r', file)
            for cand in candidates[file]:
                _debug(' * %s [%.2f]', cand['filepath'], cand['similarity'])

            # Keep only the three best matches.
            del candidates[file][3:]

        return dict(candidates)

    @staticmethod
    def _get_path_similarity(a, b, _is_junk=lambda x: x in '. -/'):
        return difflib.SequenceMatcher(_is_junk, a, b, autojunk=False).ratio()

    def _each_file(self, *paths):
        # Yield (filepath, path) tuples where the first item is a file (more
        # specifically: a non-directory) beneath the second item. The second
        # item is a path from `paths`. If there is a file path in `paths`, it is
        # yielded as both items of the tuple.
        for path in paths:
            _debug('Searching %s', path)
            if not os.path.isdir(path):
                yield (str(path), str(path))
            else:
                for root, dirnames, filenames in os.walk(path, followlinks=True):
                    for filename in filenames:
                        yield (str(os.path.join(root, filename)), str(path))

    @functools.lru_cache(maxsize=None)
    def _is_size_match(self, torrentfile, filepath):
        # Return whether a file in a torrent and an existing file are the same
        # size.
        return torrentfile.size == self._get_file_size(filepath)

    def _get_file_size(self, filepath):
        # os.path.getsize("path/to/directory") returns 4096
        if not os.path.isdir(filepath):
            try:
                return os.path.getsize(filepath)
            except OSError:
                pass
        return None

    def _create_hardlink(self, source, target):
        return self._create_link(self._hardlink_or_symlink, source, target)

    def _hardlink_or_symlink(self, source, target):
        # Try hard link and default to symlink.
        try:
            os.link(source, target)
        except OSError as e:
            if e.errno == errno.EXDEV:
                # Invalid cross-device link (`source` and `target` are on
                # different file systems)
                os.symlink(source, target)
            else:
                raise

    def _create_symlink(self, source, target):
        return self._create_link(os.symlink, source, target)

    def _create_link(self, create_link_function, source, target):
        if not os.path.exists(target):
            # Create parent directory if it doesn't exist.
            target_parent = os.path.dirname(target)
            try:
                os.makedirs(target_parent, exist_ok=True)
            except OSError as e:
                msg = e.strerror if e.strerror else e
                raise FindError(f'Failed to create directory {target_parent}: {msg}')
            else:
                # Create link.
                _debug(f'{create_link_function.__qualname__}({source!r}, {target!r})')
                try:
                    create_link_function(source, target)
                except OSError as e:
                    msg = e.strerror if e.strerror else e
                    raise FindError(f'Failed to link {source} to {target}: {msg}')
        else:
            _debug('Already exists: %r', target)

    @property
    def _temporary_directory(self):
        # Avoid illegal characters.
        name = ''.join(
            c if c in self._allowed_filename_characters else '_'
            for c in self._torrent.name
        )
        # TemporaryDirectory is a context manager that automatically deletes the
        # directory when leaving the context.
        return tempfile.TemporaryDirectory(prefix=f'{__project_name__}.{name}.')

    _allowed_filename_characters = (
        string.ascii_letters
        + string.digits
        + " ',.-"
    )



class FindError(Exception):
    """Raised by :class:`~.FindDownloadLocation` and its methods"""


class ConfigError(Exception):
    """Raised by configuration file readers"""

    def __init__(self, msg, filepath=None, line_number=None):
        if filepath and line_number:
            super().__init__(f'{filepath}@{line_number}: {msg}')
        elif filepath:
            super().__init__(f'{filepath}: {msg}')
        else:
            super().__init__(msg)



import setuptools
import re

def get_long_description():
    with open('README.rst', 'r') as f:
        return f.read()

def get_var(name):
    with open('tofipa/__init__.py') as f:
        content = f.read()
        match = re.search(rf'''^{name}\s*=\s*['"]([^'"]*)['"]''',
                          content, re.MULTILINE)
        if match:
            return match.group(1)
        else:
            raise RuntimeError(f'Unable to find {name}')

setuptools.setup(
    name=get_var('__project_name__'),
    version=get_var('__version__'),
    author=get_var('__author__'),
    author_email=get_var('__author_email__'),
    description=get_var('__description__'),
    long_description=get_long_description(),
    long_description_content_type='text/x-rst',
    url=get_var('__homepage__'),
    packages=setuptools.find_packages(),
    classifiers=[
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
    ],
    python_requires='>=3.8',
    install_requires=[
        'aiobtclientapi==1.*',
        'torf==4.*',
        'pyxdg==0.*',
    ],
    entry_points={'console_scripts': ['tofipa = tofipa:cli']},
)
